from itertools import cycle
import numpy as np
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d
from tc_python import *

"""
Simulation of the homogenization heat treatment of an alloy from the as-cast state.

As a first step, the microsegregation after solidification is simulated using a Scheil-calculation, then
as a second step the homogenization is simulated using a diffusion simulation.
"""


thermodynamic_database = "ALDEMO"
mobility_database = "MALDEMO"
dependent_element = "Al"
composition = {"Cu": 3.5, "Si": 6.0}  # in wt-%
num_interp_points = 100
alloy = "Alloy 319.2"
dendrite_arm_spacing = 22  # in micrometers
homogenization_temperature = 515  # in degree C
homogenization_time = 8.0  # in hours
FIRST_TIMEPOINT_WITH_PHASES = 1e-5  # to get the initial phase distribution close to t=0


with TCPython() as session:
    elements = list(composition.keys())
    system = (session
              .set_cache_folder(os.path.basename(__file__) + "_cache")
              .select_thermodynamic_and_kinetic_databases_with_elements(thermodynamic_database, mobility_database,
                                                                        [dependent_element] + elements)
              .deselect_phase("GAS")
              .get_system())

    calc = (system.with_scheil_calculation()
            .set_composition_unit(CompositionUnit.MASS_PERCENT))
    for element in composition:
        calc.set_composition(element, composition[element])
    result = calc.calculate()

    # obtain all phases formed in the system
    scheil_curve = result.get_values_grouped_by_stable_phases_of(ScheilQuantity.mole_fraction_of_all_solid_phases(),
                                                                 ScheilQuantity.temperature())
    phases_in_calc = set()
    for label in scheil_curve:
        splitted_label = label.split("+")
        splitted_label = set([x.strip() for x in splitted_label])
        for phase in splitted_label:
            phases_in_calc.add(phase)

    __, reference_temps = result.get_values_of(ScheilQuantity.mole_fraction_of_all_solid_phases(),
                                               ScheilQuantity.temperature())

    # calculate the solid concentration along the solidification
    total_solid_fracs = None
    moles_of_solid_in_slice = None
    moles_of_element_in_slice = dict()
    for phase in phases_in_calc:
        if system.get_phase_object(phase).get_name().startswith("LIQUID"):
            continue

        total_solid_fracs, mole_frac_of_this_solid_phase = \
            result.get_values_of(ScheilQuantity.mole_fraction_of_all_solid_phases(),
                                 ScheilQuantity.mole_fraction_of_a_solid_phase(phase))
        moles_of_phase_in_slice = np.diff(mole_frac_of_this_solid_phase)
        moles_of_solid_in_slice = np.diff(total_solid_fracs)

        for element in elements:
            print("Calculating for element {}, phase {}".format(element, phase))
            current_phase_fracs, concentration = \
                result.get_values_of(ScheilQuantity.mole_fraction_of_all_solid_phases(),
                                     ScheilQuantity.composition_of_phase_as_mole_fraction(phase, element))

            if element not in moles_of_element_in_slice:
                moles_of_element_in_slice[element] = 0

            moles_of_element_in_phase_in_slice = np.multiply(concentration[1:], moles_of_phase_in_slice)
            moles_of_element_in_slice[element] += moles_of_element_in_phase_in_slice

    average_solid_compo_in_slice = dict()
    for element in elements:
        average_solid_compo_in_slice[element] = moles_of_element_in_slice[element] / moles_of_solid_in_slice

    # calculate the liquid concentration along the solidification
    liquid_compo_along_solid = dict()
    for element in elements:
        current_phase_fracs, liquid_concentration = \
            result.get_values_of(ScheilQuantity.mole_fraction_of_all_solid_phases(),
                                 ScheilQuantity.composition_of_phase_as_mole_fraction("LIQUID", element))
        liquid_compo_along_solid[element] = liquid_concentration

    # interpolation
    average_solid_compo_in_slice_interp = dict()
    for element in elements:
        f = interp1d(total_solid_fracs,
                     list(average_solid_compo_in_slice[element]) + [average_solid_compo_in_slice[element][-1]],
                     fill_value="extrapolate")

        x_interp = f(np.linspace(0, 1, num_interp_points))
        x_interp[np.isnan(x_interp)] = 0
        x_interp[x_interp < 0] = 0
        average_solid_compo_in_slice_interp[element] = x_interp

    # homogenization simulation
    concentration_profiles = PointByPointGrid(unit_enum=Unit.MOLE_FRACTION)
    for i in range(0, num_interp_points):
        gridpoint = GridPoint(i / (num_interp_points - 1) * (dendrite_arm_spacing * 1e-6 / 2))
        for element in elements:
            gridpoint.add_composition(element, average_solid_compo_in_slice_interp[element][i])
        concentration_profiles.add_point(gridpoint)

    dendrite_region = (Region("Solid")
                       .with_point_by_point_grid_containing_compositions(concentration_profiles))

    for phase in phases_in_calc:
        if not system.get_phase_object(phase).get_name().startswith("LIQUID"):
            dendrite_region.add_phase(phase)

    dictra = (system.with_isothermal_diffusion_calculation()
              .set_temperature(homogenization_temperature + 273.15)
              .set_simulation_time(homogenization_time * 3600)
              .add_region(dendrite_region)
              .with_solver(Solver.automatic()))

    homo_results = dictra.calculate()

    # plot results
    plt.figure(1)
    plt.title("Alloy: {}".format(alloy) + ", Secondary arm spacing = {} \N{MICRO SIGN}m".
              format(dendrite_arm_spacing))
    for (label, scheil_stable_phases) in scheil_curve.items():
        plt.plot(scheil_stable_phases.x, scheil_stable_phases.y, label=label)
    plt.xlabel("Fraction solid phase")
    plt.ylabel("Temperature (K)")
    plt.legend(loc="best")

    plt.figure(2)
    plt.title("Alloy: {}".format(alloy) + ", Secondary arm spacing = {} \N{MICRO SIGN}m, {} \N{DEGREE SIGN}C".
              format(dendrite_arm_spacing, homogenization_temperature))
    linestyles = cycle(['-', '--', '-.', ':'])
    for time in [FIRST_TIMEPOINT_WITH_PHASES, homogenization_time / 2, homogenization_time]:
        linestyle = next(linestyles)
        colors = cycle(["b", "r", "g", "m"])
        for element in elements:
            color = next(colors)
            distance, x_solid_homogenized = homo_results.get_mole_fraction_of_component_at_time(element, time * 3600)
            if time == FIRST_TIMEPOINT_WITH_PHASES:
                time_label = 0
            else:
                time_label = time
            plt.plot(np.array(distance) * 1e6, x_solid_homogenized,
                     ls=linestyle, label="{} after {} h".format(element, time_label), color=color)

    plt.xlabel("distance (\N{MICRO SIGN}m)")
    plt.ylabel("concentration (mole-fraction)")
    plt.legend(loc="best")
    plt.show()

    plt.figure(3)
    plt.title("Alloy: {}".format(alloy) + ", Dendrite arm spacing = {} \N{MICRO SIGN}m, {} \N{DEGREE SIGN}C".
              format(dendrite_arm_spacing, homogenization_temperature))
    linestyles = cycle(['-', '--', '-.', ':'])
    for time in [FIRST_TIMEPOINT_WITH_PHASES, homogenization_time / 2, homogenization_time]:
        linestyle = next(linestyles)
        colors = cycle(["b", "r", "g", "m"])
        for phase in phases_in_calc:
            if system.get_phase_object(phase).get_name().startswith("LIQUID"):
                continue
            color = next(colors)

            distance, phase_frac = homo_results.get_mole_fraction_of_phase_at_time(phase, time * 3600)
            if time == FIRST_TIMEPOINT_WITH_PHASES:
                time_label = 0
            else:
                time_label = time
            plt.plot(np.array(distance) * 1e6, np.array(phase_frac) * 100,
                     ls=linestyle, label="{} after {} h".format(phase, time_label), color=color)

    plt.xlabel("distance (\N{MICRO SIGN}m)")
    plt.ylabel("phase fraction (vol-%)")
    plt.legend(loc="best")
    plt.show()
